/*
 * Copyright (c) 2007 Henric Jungheim <software@henric.info>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include "stdafx.h"

using namespace System;
using namespace System::Security::Cryptography;

namespace PlatformMath {
ref class RandomSFMT;

public ref class Status abstract sealed 
{
	static bool haveSSE2;
	static RandomSFMT^ rng;

	static Status()
	{
		haveSSE2 = IsProcessorFeaturePresent(PF_XMMI64_INSTRUCTIONS_AVAILABLE) ? true : false;
	}
public:
	static property bool HasSSE2 { bool get() { return haveSSE2; } }
	static property double RandomValue { double get(); }
};

	public ref class RandomSFMT abstract
	{
	protected:
		array<double>^ buffer;
		int index;
		int length;

		virtual void FillBuffer() = 0;
		static Object^ globalLock = gcnew Object();
		static RNGCryptoServiceProvider rng;
		static RandomSFMT^ fastRng;

		static array<Byte>^ GetRandomArray(int size)
		{
			array<Byte>^ data = gcnew array<Byte>(size);
			rng.GetBytes(data);
			return data;
		}
	public:
		RandomSFMT()
		{
			buffer = gcnew array<double>(SFMT_N * 2 + 2 + 1);
			index = length = buffer->Length;
		}
		virtual double NextDouble()
		{
			if (index >= length)
				FillBuffer();

			return buffer[index++];
		}

		virtual void Reseed() = 0;

		property String^ Name
		{
			String^ get()
			{
				return SFMT_IDSTR;
			}
		}

		static RandomSFMT^ CreateRandom();
	};

	ref class SSE2_dSFMT sealed : public RandomSFMT
	{
	protected:
		virtual void FillBuffer() override
		{
			pin_ptr<double> p = &buffer[0];
			double* porig = p;

			intptr_t ip0 = reinterpret_cast<intptr_t>(p);
			intptr_t ip = ip0;

			double* p2 = porig;
			length = buffer->Length;

			/* Force 16 byte alignment */
			if (ip & 0x0f)
			{
				/* Round up */
				ip |= 0x0f;
				++ip;

				ptrdiff_t offset = ip - ip0;
				while (offset > 0) {
					--length;
					offset -= sizeof(double);
				}

				p2 = reinterpret_cast<double *>(ip);
			}

#ifdef _DEBUG
			if (0x0f & reinterpret_cast<intptr_t>(p2))
				throw gcnew Exception("Internal error");
#endif

			/* Length must be a power-of-two */
			if (length & 1)
				--length;

			{ /* Scope */
				msclr::lock lock(globalLock);

				fill_array_close_open_sse2(p2, length);
			}

			index = 0;

			/* Realign the doubles to match buffer */
			if (ip0 & 0x07)
			{
				memmove(p, p2, length * sizeof(double));
			}
			else
			{
				while (ip > ip0)
				{
					ip -= sizeof(double);
					++index;
				}
			}
		}

	public:
		SSE2_dSFMT()
		{
			Reseed();
		}

		virtual void Reseed() override
		{
			length = index = 0;
			array<Byte>^ initData = GetRandomArray(SFMT_N64 * sizeof(int) + sizeof(int) - 1);

			pin_ptr<Byte> p = &initData[0];

			intptr_t ip = (intptr_t)p;
			while (ip & (sizeof(int) - 1))
				++ip;

			msclr::lock lock(globalLock);
			init_by_array_sse2(reinterpret_cast<uint32_t *>(ip), SFMT_N64);
		}
#if 0
		virtual void Reset(int seed) override
		{
			length = 0;
			msclr::lock lock(globalLock);
			init_gen_rand_sse2(seed);
		}
#endif // 0
	};

#if !defined(_M_X64)
	ref class dSFMT sealed : public RandomSFMT
	{
	protected:
		virtual void FillBuffer() override
		{
			index = 0;
			length = buffer->Length & ~1;

			pin_ptr<double> p = &buffer[0];
			msclr::lock lock(globalLock);

			fill_array_close_open(p, length);
		}
	public:
		dSFMT()
		{
			Reseed();
		}

		virtual void Reseed() override
		{
			length = index = 0;

			array<Byte>^ initData = GetRandomArray(SFMT_N64 * sizeof(int) + sizeof(int) - 1);

			pin_ptr<Byte> p = &initData[0];

			intptr_t ip = (intptr_t)p;
			while (ip & (sizeof(int) - 1))
				++ip;

			msclr::lock lock(globalLock);
			init_by_array(reinterpret_cast<uint32_t *>(ip), SFMT_N64);
		}

#if 0
		virtual void Reset(int seed) override
		{
			length = 0;
			msclr::lock lock(globalLock);
			init_gen_rand(seed);
		}
#endif // 0
	};
#endif // !defined(_M_X64)

	RandomSFMT^ RandomSFMT::CreateRandom()
	{
#if defined(_M_X64)
		return gcnew SSE2_dSFMT();
#else
		if (Status::HasSSE2)
			return gcnew SSE2_dSFMT();
		else
			return gcnew dSFMT();
#endif // defined(_M_X64)
	}

	double Status::RandomValue::get()
	{
		RandomSFMT^ newRng = nullptr;
		if (nullptr == rng)
		{
			newRng = RandomSFMT::CreateRandom();
			System::Threading::Interlocked::CompareExchange<RandomSFMT^>(rng, newRng, nullptr);
		}

		newRng = rng;
		if (nullptr == newRng)
			throw gcnew Exception("Internal error");

		msclr::lock lock(newRng);
		return newRng->NextDouble();
	}
}
